//---------------------------------------------------------------------------

#include <vcl.h>
#pragma hdrstop

#include "HistogramMatching0.h"
#include <stdio.h>
//---------------------------------------------------------------------------
#pragma package(smart_init)
#pragma resource "*.dfm"
TForm1 *Form1;
//---------------------------------------------------------------------------
__fastcall TForm1::TForm1(TComponent* Owner)
        : TForm(Owner)
{
}
//---------------------------------------------------------------------------
const int RED   = 2;
const int GREEN = 1;
const int BLUE  = 0;
const int GRAY  = 4;

class ImageTransfer
{ private:
    Graphics::TBitmap *mBmp;
    float * mDisModel;
  public:
    ImageTransfer() { mBmp=NULL; mDisModel=NULL; }
    ~ImageTransfer() { if ( mDisModel!=NULL ) delete [] mDisModel; }
    void assignBitmap(Graphics::TBitmap *bmp )
    { //if ( mBmp!=NULL ) delete bmp;
      mBmp=bmp;
      mBmp->PixelFormat = pf24bit;
    }
    void assignDisModel(float *dm)
    { mDisModel = dm; }
    int getWidth()
    { if ( mBmp!=NULL ) return(mBmp->Width);
      else return(-1);
    }
    int getHeight()
    { if ( mBmp!=NULL ) return(mBmp->Height);
      else return(-1);
    }
    void keepColor(int color);
    float * getGrayImage();
    void toRImage() { keepColor(RED);   }
    void toGImage() { keepColor(GREEN); }
    void toBImage() { keepColor(BLUE);  }
    void toGrayImage();
    void toHistogramize(bool withModel=false);
    void saveDisModel(char *filename, int type=GRAY);
    void loadDisModel(char *filename);
    void showDisModel(TCanvas *canvas, int x0, int y0);
    void saveToFile(char * fileName)
      { if ( mBmp!=NULL )
          { mBmp->PixelFormat = pf24bit;
            mBmp->SaveToFile(fileName);
          }
      }
};
void ImageTransfer::keepColor(int color)
{ Byte *ptr, *value;
  int  x, y, c[2];

  if ( color==RED )
    { c[0]=GREEN; c[1]=BLUE;
    }
  else if ( color==GREEN )
    { c[0]=RED; c[1]=BLUE;
    }
  else { c[0]=RED; c[1]=GREEN;
       }
  mBmp->PixelFormat = pf24bit;
  for (int y=0; y<mBmp->Height; y++)
     { ptr = (Byte *) mBmp->ScanLine[y];
       for (int x=0; x<mBmp->Width; x++)
          { ptr[ 3*x+c[0] ] = 0;
            ptr[ 3*x+c[1] ] = 0;
          }
     }
}
float * ImageTransfer::getGrayImage()
{ float *image, gray;
  Byte *ptr, value;
  int index, idxImage;

  mBmp->PixelFormat = pf24bit;
  idxImage = 0L;
  image = new float [mBmp->Height * mBmp->Width];
  if ( image==NULL )
    { ShowMessage("Memory allocation error !!");
      return NULL;
    }
  for (int y = 0; y < mBmp->Height; y++)
    { ptr = (Byte *)mBmp->ScanLine[y];
      index=0;
      for (int x = 0; x < mBmp->Width; x++)
         { gray = 0.299*ptr[index+RED]  +
                  0.587*ptr[index+GREEN]+
                  0.114*ptr[index+BLUE];
           image[idxImage++] = gray;
           index += 3;
         }
    }
  return (float*) image;
}
void ImageTransfer::toGrayImage()
{ float gray;
  Byte *ptr, value;
  int index;

  mBmp->PixelFormat = pf24bit;
  for (int y = 0; y < mBmp->Height; y++)
    { ptr = (Byte *)mBmp->ScanLine[y];
      index=0;
      for (int x = 0; x < mBmp->Width; x++)
         { gray = 0.297*ptr[index+RED]  +
                  0.589*ptr[index+GREEN]+
                  0.114*ptr[index+BLUE];
           value = (gray);
           ptr[index+  RED] = value;
           ptr[index+GREEN] = value;
           ptr[index+ BLUE] = value;
           index += 3;
         }
    }
}
void ImageTransfer::showDisModel(TCanvas *canvas, int x0, int y0)
{ int gray, height=80;

  if ( mDisModel==NULL ) return;
  canvas->Rectangle(x0,y0,x0+257, y0+height);
  for (gray=0; gray<256; gray++)
     { canvas->MoveTo(x0+gray, y0+height);
       canvas->LineTo(x0+gray, y0+height-height*mDisModel[gray]);
     }
}
void ImageTransfer::loadDisModel(char *filename)
{ FILE *fp;

  fp = fopen(filename,"rb");
  if ( fp==NULL )
    ShowMessage("File open error !");
  else
    { if ( mDisModel==NULL )
        mDisModel = new float [256];
      fread(mDisModel, sizeof(float)*256, 1, fp);
      fclose(fp);
    }
}
void ImageTransfer::saveDisModel(char *filename, int type)
{ int row, col, mWidth, mHeight, clr;
  float *image, hist[256], factor;
  BYTE *ptr;
  unsigned long index=0L;

  if ( mBmp==NULL )
    assignBitmap(Form1->Image1->Picture->Bitmap);
  //toGrayImage();
  mWidth  = mBmp->Width;
  mHeight = mBmp->Height;
  if ( type==GRAY )
    { toGrayImage();
      clr = 0;
    }
  else clr = type;
  // read image data into image
  image = new float[mWidth*mHeight];
  index = 0L;
  for (row=0; row<mHeight; ++row)
    { ptr = (BYTE *) mBmp->ScanLine[row];
      for (col=0; col<mWidth; ++col)
         {
           image[index+col] = ptr[3*col+clr];
         }
      index += mWidth;
    }
  for (row=0; row<256; row++) hist[row]=0.0;
  index = 0L;
  for (row=0; row<mHeight; ++row)
     { for (col=0; col<mWidth; ++col)
          { hist[ (int)image[index++] ]++;
          }
     }
  factor = 1./((float)mWidth*mHeight);
  hist[0] *= factor; // the Pr(r) obtained
  for (row=1; row<256; row++)
     { hist[row] *= factor;
       hist[row] += hist[row-1];
       if ( hist[row]>1.0 ) hist[row]=1.0;
     }
  FILE *fp;
  fp = fopen(filename,"wb");
  try { if ( fp==NULL )
          ShowMessage("File open error !");
        else
          { fwrite(hist, sizeof(float)*256, 1, fp);
            fclose(fp);
          }
      }
  __finally
    { delete [] image;
    }
}
void ImageTransfer::toHistogramize(bool withModel)
{ int row, col, mWidth, mHeight;
  float delta, *image=NULL, hist[256], factor;
  BYTE *ptr;
  unsigned long index=0L;

  if ( withModel && mDisModel==NULL )
    { ShowMessage("Select the Accumulated-Distribution-Model file first !!");
      Form1->OpenDialog1->Filter = "*.dm|*.dm";
      if ( Form1->OpenDialog1->Execute() )
        { loadDisModel( Form1->OpenDialog1->FileName.c_str() );
          showDisModel(Form1->Canvas, 0,0);
        }
      else return;
    }

  mWidth  = mBmp->Width;
  mHeight = mBmp->Height;
  // read image data into image
  image = getGrayImage();
  // get the histogram of gray style
  for (row=0; row<256; row++) hist[row]=0.0;
  index = 0L;
  for (row=0; row<mHeight; ++row)
     { for (col=0; col<mWidth; ++col)
          { hist[ (int)image[index++] ]++;
          }
     }

  factor = 1./((float)mWidth*mHeight);
  hist[0] *= factor; // form Pr(r) to T(r)
  for (row=1; row<256; row++)
     { hist[row] *= factor;
       hist[row] += hist[row-1];
       if ( hist[row]>1 ) hist[row]=1.0;
     }
  // from Tr(r) equal to s[.]
  int  gray, z;
  float pz, level;
  factor = 1.0/256;

  if ( withModel )  // hist[.] is the s[.]
    { z=0;
      for (gray=0; gray<256; gray++)
         { while ( hist[gray]>mDisModel[z] )
             { if ( ++z>255 )
                 { z = 255;
                   break;
                 }
             }
      //dis[z]=hist[gray];
           hist[gray] = z;

         }
    }
  else
    { for (gray=0; gray<256; gray++) hist[gray]=gray;
    }

  index=0L;
  // save the image to mBmp
  BYTE color;
  index = 0L;
  float scale, r,g,b;
  for (row=0; row<mHeight; ++row)
    { ptr = (BYTE *) mBmp->ScanLine[row];
      for (col=0; col<mWidth; ++col)
         { color = hist[ (int)image[index+col] ];
           if ( image[index+col]==0.0 ) scale=1.0;
           else scale = (float)color/image[index+col];
           r = ptr[3*col+RED  ] * scale;
           g = ptr[3*col+GREEN] * scale;
           b = ptr[3*col+BLUE ] * scale;
           if ( r>255 ) r=255;
           if ( g>255 ) g=255;
           if ( b>255 ) b=255;

           ptr[3*col+RED  ] = r; //scale; //color;
           ptr[3*col+GREEN] = g; //scale; //color;
           ptr[3*col+BLUE ] = b; //scale;  // color;
         }
      index += mWidth;
    }
  
  if ( image!=NULL ) delete [] image;
}

void __fastcall TForm1::Button1Click(TObject *Sender)
{
  ImageTransfer imgTransfer;

  imgTransfer.assignBitmap(Image1->Picture->Bitmap);

  Form1->Canvas->Draw(Image1->Picture->Width+2, 1, Image1->Picture->Graphic);
  imgTransfer.loadDisModel("image1.dm");

  imgTransfer.toHistogramize(true);
}
//---------------------------------------------------------------------------
void __fastcall TForm1::Button2Click(TObject *Sender)
{
  SaveDialog1->Filter = "*.dm|*.dm";
  SaveDialog1->DefaultExt = "dm";
  if ( SaveDialog1->Execute() )
    { ImageTransfer imgTransfer;
      imgTransfer.assignBitmap(Image1->Picture->Bitmap);
      imgTransfer.saveDisModel(SaveDialog1->FileName.c_str());
    }
}
//---------------------------------------------------------------------------
